unit uFunc;

{$INCLUDE 'h2lrf.inc'}

interface

uses
  SysUtils, Classes,
  {$IFDEF FPC}
    Dos,
  {$ELSE}
    Windows, FileCtrl,
  {$ENDIF}
  IniFiles, uPCRE, uFiles, uCompatibility;

{$IFDEF WIN32}
  const EOLN = #13#10;
{$ELSE}
  const EOLN = #10;
{$ENDIF}

const
  META_ANY = '*';
  META_ANY_PREFIX = META_ANY + ':';
  META_INCLUDE = '@include';

type
  TConvertOptions = record
    ConverterPath: string;          // Location of the converter
    ConverterCommandLine: string;   // Command line structure for the converter. Uses %%, %PARAMS%, %INPUT%, %OUTPUT%
    InputDir: string;               // Where to start looking for source book files
    InputWildCard: string;          // Files matching this wildcard will be converted
    InputRecursive: boolean;        // Scan subdirectories of InputDir?
    InputMetaPrefix: string;        // Prefix for H2LRF's <meta name="*"> in source files
    OutputDir: string;              // Where the converted files will be stored
    OutputRecursive: boolean;       // Replicate directory structure from Input for converted files
    OutputExtension: string;        // Extensions of the converted files
    ErrorLogFileName: string;       // Error messages will be saved into this file
    AbsoluteLocalUrlPrefix: string; // Needed for Calibre 0.6: prefix to use for absolute local URLs; standard="file:///", Calibre 0.6="" 
    end;

function SetTarget(const ATarget: string; var Options: TConvertOptions): boolean;
function Convert(const Options: TConvertOptions): boolean;
function DefaultOptions: TConvertOptions;

implementation

const
  META_ANY_PREFIX_LENGTH = Length(META_ANY_PREFIX);

const
  ANY_FILE = {$IFDEF FPC} '*' {$ELSE} '*.*' {$ENDIF};

const
  DEFAULT_TARGET = 'lrf';
  DEFAULT_COMMAND_LINE = '%PARAMS% --output="%OUTPUT%" "%INPUT%"';

type
  Eh2lrfError = class(Exception);

{$IFDEF CLEANUP}
var
  CleanUp: TStringList;
{$ENDIF}

var
  MetaPrefix: string;
  MetaPrefixLength: integer;
  AbsoluteLocalUrlPrefix: string;

function OpenIniFile: TIniFile;
  // Open application's INI file
var Name: string;
begin
  Name := ChangeFileExt(ParamStr(0), '.ini');
  if FileExists(Name) then
    Result := TIniFile.Create(Name)
  else
    Result := nil;
end;

function DefaultOptions: TConvertOptions;
var Target: string;
begin
  Target := DEFAULT_TARGET;
  Result.ConverterPath := '';
  Result.ConverterCommandLine := DEFAULT_COMMAND_LINE;
  Result.InputDir := '.\';
  Result.InputWildCard := '';
  Result.InputRecursive := False;
  Result.InputMetaPrefix := Target + ':';
  Result.OutputDir := GetCurrentDir; //'.\';
  Result.OutputRecursive := False;
  Result.OutputExtension := '.' + Target;
  Result.ErrorLogFileName := '';
  Result.AbsoluteLocalUrlPrefix := 'file:///';
  try
    with OpenIniFile do
      try
        Target := ReadString('Config', 'DefaultTarget', Target);
        Result.AbsoluteLocalUrlPrefix := ReadString('Config', 'AbsoluteLocalUrlPrefix', Result.AbsoluteLocalUrlPrefix);
      finally
        Free;
        end;
  except
    on Exception do
      ;
    end;
  SetTarget(Target, Result);
end;

function SetTarget(const ATarget: string; var Options: TConvertOptions): boolean;
  // Get filename of the application used to convert files
begin
  Result := False;
  if ATarget <> '' then
    with OpenIniFile do
      try
        Options.ConverterPath := ReadString(ATarget, 'Converter', ReadString('Config', 'Converter', ''));
        Options.ConverterCommandLine := ReadString(ATarget, 'CommandLine', ReadString('Config', 'CommandLine', DEFAULT_COMMAND_LINE));
        Options.InputMetaPrefix := ReadString(ATarget, 'Prefix', LowerCase(ATarget) + ':');
        Options.OutputExtension := ReadString(ATarget, 'Extension', '.' + LowerCase(ATarget));
        Result := (Options.ConverterPath <> '') and FileExists(Options.ConverterPath);
      finally
        Free;
        end;
end;

function Execute(const Path, CommandLine: string; out ResultCode: DWORD): boolean;
  // Execute an external application, wait until it terminates and return the result code
{$IFDEF FPC}
{$ELSE}
  var StartupInfo: TStartupInfo;
      ProcessInfo: TProcessInformation;
{$ENDIF}
begin
  Writeln('EXEC: ' + Path + ' ' + CommandLine);
  Flush(Output);
  {$IFDEF FPC}
    Exec(Path, CommandLine);
    if DosError = 0 then
      begin
      Result := True;
      ResultCode := DosExitCode;
      end
    else
      Result := False;
  {$ELSE}
    FillChar(StartupInfo, SizeOf(TStartupInfo), 0);
    StartupInfo.cb := SizeOf(StartupInfo);
    if CreateProcess(nil, PChar(Path + ' ' + CommandLine), nil, nil, TRUE, 0, nil, nil, StartupInfo, ProcessInfo) then
      begin
      repeat
        GetExitCodeProcess(ProcessInfo.hProcess, ResultCode);
        if ResultCode <> STILL_ACTIVE then
          Sleep(200);
      until ResultCode <> STILL_ACTIVE;
      Result := True;
      end
    else
      Result := False;
  {$ENDIF}
end;

function GetLastErrorStr(LastError: DWORD): string;
  // Get system error string
{$IFDEF FPC}
{$ELSE}
  const BUFSIZE = 2048;
  var Buf: array[0..BUFSIZE] of char;
{$ENDIF}
begin
  {$IFDEF FPC}
    case LastError of
       0: Result := 'OK.';
       2: Result := 'File not found.';
       3: Result := 'Path not found.';
       5: Result := 'Access denied.';
       6: Result := 'Invalid handle.';
       8: Result := 'Not enough memory.';
      10: Result := 'Invalid environment.';
      11: Result := 'Invalid format.';
      18: Result := 'No more files.';
      else Result := '?';
      end;
  {$ELSE}
    if FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, LastError, 0, @Buf[0], BUFSIZE, nil) = 0 then
      Result := '?'
    else
      Result := Buf;
  {$ENDIF}
end;

function SystemTempDir: string;
  // Return temp directory
{$IFDEF FPC}
{$ELSE}
  var Buf: array[0..MAX_PATH] of Char;
      n: integer;
{$ENDIF}
begin
  {$IFDEF FPC}
    Result := GetTempDir(False);
    if Result = '' then
      Result := GetTempDir(True);
  {$ELSE}
    n := GetTempPath(Length(Buf), Buf);
    if (n > 0) and (n < Length(Buf)) then
      Result := StrPas(Buf)
    else
      Result := '';
  {$ENDIF}
end;

function SystemTempFile(const Dir, Prefix: string): string;
  // Return temp file name
{$IFDEF FPC}
{$ELSE}
  var n: integer;
      Buf: array[0..MAX_PATH] of Char;
{$ENDIF}
begin
  {$IFDEF FPC}
    Result := GetTempFileName(Dir, Prefix);
  {$ELSE}
    n := GetTempFileName(PChar(Dir), PChar(Prefix), 0, Buf);
    if n <> 0 then
      Result := string(Buf)
    else
      Result := '';
  {$ENDIF}
end;

{$IFDEF CLEANUP}
procedure RemoveDirectory(const Dir: string);
  // Delete directory and all of its subdirectories
var SR: TSearchRec;
    OK: integer;
begin
  if (Dir <> '') and DirectoryExists(Dir) then
    begin
    OK := SysUtils.FindFirst(Dir + '\' + ANY_FILE, faAnyFile, SR);
    try
      while OK = 0 do
        begin
        if Longbool(SR.Attr and faDirectory) then
          begin
          if (SR.Name <> '.') and (SR.Name <> '..') then
            RemoveDirectory(Dir + '\' + SR.Name);
          end
        else
          SysUtils.DeleteFile(Dir + '\' + SR.Name);
        OK := SysUtils.FindNext(SR);
        end;
    finally
      SysUtils.FindClose(SR);
      end;
    RemoveDir(Dir);
    end;
end;

procedure DoCleanUp;
  // Remove all temporary directories
var i: integer;
begin
  for i := 0 to Pred(CleanUp.Count) do
    RemoveDirectory(CleanUp[i]);
  CleanUp.Clear;
end;
{$ENDIF}

procedure ReportError(E: Exception; const Log, FileName: string);
  // Report error to the user
var s: string;
    T: TextFile;
begin
  s := Format('%sERROR %s WHILE CONVERTING %s:%s%s%s', [EOLN, E.ClassName, FileName, EOLN, E.Message, EOLN]);
  Writeln(s);
  if Log <> '' then
    begin
    AssignFile(T, Log);
    try
      if FileExists(Log) then
        Append(T)
      else
        Rewrite(T);
      try
        Writeln(T, s);
      finally
        CloseFile(T);
        end;
    except
      on Exception do
        ;
      end;
    end;
end;

function DecodeHtmlString(const s: string): string;
  // Decode some HTML entities
begin
  Result := s;
  Result := StringReplace(Result, '&lt;', '<', [rfReplaceAll, rfIgnoreCase]);
  Result := StringReplace(Result, '&gt;', '>', [rfReplaceAll, rfIgnoreCase]);
  Result := StringReplace(Result, '&quot;', '"', [rfReplaceAll, rfIgnoreCase]);
  Result := StringReplace(Result, '&amp;', '&', [rfReplaceAll, rfIgnoreCase]);
end;

function EscapeQuotes(const s: string): string;
begin
  // Escape quotes for inclusion in command line
  Result := AnsiQuotedStr(s, '"');
end;

function ReplaceInString(const s, RealFileName: string): string;
  // Replace certain symbols with their respective values:
  // '%%' => '%'
  // '%f' => full filename (drive, path, filename, extension)
  // '%d' => drive
  // '%p' => path, including drive
  // '%n' => just filename without extension
  // '%x' => extension
  // '%e' => ''
  // '%r' => remove entirely

  procedure InsertIntoResult(const Co: string; var Res: string; var Pos: integer);
    begin
      Insert(Co, Res, Pos);
      Inc(Pos, Length(Co));
    end;

var i, j, n: integer;
begin
  Result := s;
  i := 1;
  j := 1;
  n := Length(s);
  while i <= n do
    begin
    if (s[i] = '%') and (i < n) then
      case s[Succ(i)] of
        '%': // '%%' => '%'
             begin
             Delete(Result, Succ(j), 1);
             end;
        'f': // '%f' => cele jmeno souboru
             begin
             Inc(i);
             Delete(Result, j, 2);
             InsertIntoResult(RealFileName, Result, j);
             end;
        'd': // '%d' => jmeno disku
             begin
             Inc(i);
             Delete(Result, j, 2);
             InsertIntoResult(ExtractFileDrive(RealFileName), Result, j);
             end;
        'p': // '%p' => cesta k souboru vcetne jmena disku
             begin
             Inc(i);
             Delete(Result, j, 2);
             InsertIntoResult(ExtractFilePath(RealFileName), Result, j);
             end;
        'n': // '%n' => jen jmeno souboru bez pripony
             begin
             Inc(i);
             Delete(Result, j, 2);
             InsertIntoResult(ChangeFileExt(ExtractFileName(RealFileName), ''), Result, j);
             end;
        'x': // '%x' => pripona souboru
             begin
             Inc(i);
             Delete(Result, j, 2);
             InsertIntoResult(ExtractFileExt(RealFileName), Result, j);
             end;
        'e', // '%e' => ''
        'r': // '%r' => fallback pro pripad, ze by nekde zustalo
             begin
             Inc(i);
             Delete(Result, j, 2);
             end;
        end;
    Inc(i);
    end;
end;

function ValidMetaPrefix(var Name: string): boolean;
  // Test, if this is a wanted META
begin
  Result := AnsiCompareText(Copy(Name, 1, MetaPrefixLength), MetaPrefix) = 0;
  if Result then
    Name := Copy(Name, MetaPrefixLength+1, MaxInt)
  else
    begin
    Result := AnsiCompareText(Copy(Name, 1, META_ANY_PREFIX_LENGTH), META_ANY_PREFIX) = 0;
    if Result then
      Name := Copy(Name, META_ANY_PREFIX_LENGTH+1, MaxInt);
    end;
end;

function ValidMetaInclude(Name: string): boolean;
  // Test, if this is a wanted META INCLUDE
begin
  Result := AnsiCompareText(Name, META_INCLUDE) = 0;
  if not Result then
    if ValidMetaPrefix(Name) then
      Result := AnsiCompareText(Name, META_INCLUDE) = 0;
end;

function ReadParametersFromFile(const FileContent, RealFileName: string; Params, Values: TStringList): boolean;
  // Read command line parameters for the conversion process (all those <meta name="lrf:..." content="..."> tags)
var Meta: TRegExp;
    Name, Value: string;
    Index: integer;
begin
  Result := True;
  Meta := RegExCreate('<meta\s+name=(?P<QUOTE1>["'']?)(?P<NAME>[^>]*?)(?P=QUOTE1)\s+content=(?P<QUOTE2>["'']?)(?P<VALUE>[^>]*?)(?P=QUOTE2).*?>');
  try
    if Meta.Match(FileContent) then
      repeat
        Name := DecodeHtmlString(Meta.SubexpressionByName('NAME'));
        Value := DecodeHtmlString(Meta.SubexpressionByName('VALUE'));
        if ValidMetaPrefix(Name) then
          begin
          Index := Params.IndexOf(Name);
          if AnsiCompareText(Value, '%r') = 0 then
            begin
            if Index >= 0 then
              begin
              Params.Delete(Index);
              Values.Delete(Index);
              end;
            end
          else
            begin
            Value := ReplaceInString(Value, RealFileName);
            if Index >= 0 then
              Values[Index] := Value
            else
              begin
              Params.Add(Name);
              Values.Add(Value);
              end;
            end;
          //Writeln(Format('"%s"="%s"', [Name, Value]));
          end;
      until not Meta.MatchAgain;
  finally
    Meta.Free;
    end;
end;

function RelativePathToFullPath(const BaseFileName, FileName: string): string;
  // Convert relative path to fully qualified absolute path
var FN, Para: string;
    ix: integer;
begin
  ix := Pos('#', FileName);
  if ix <= 0 then
    begin
    FN := FileName;
    Para := '';
    end
  else
    begin
    FN := Copy(FileName, 1, Pred(ix));
    Para := Copy(FileName, ix, MaxInt);
    end;
  if FN = '' then
    Result := ''
  else if ExtractFileDrive(FN) <> '' then
    Result := ExpandFileName(FN)
  else if CharInSet(FN[1], ['\', '/']) then
    Result := ExtractFileDrive(ExpandFileName(BaseFileName)) + FN
  else
    Result := ExtractFilePath(ExpandFileName(BaseFileName)) + FN;
  Result := Result + Para;
end;

function ReadFileWithAllIncludes(const FileName: string; out Encoding: TFileEncoding): string;
  // Read file content and all included files
var Includes, Paths: TRegExp;
    Content, Name, Value, Path: string;
    DummyEncoding: TFileEncoding;
    ix: integer;
begin
  // Read the file itself
  Content := LoadFileIntoString(FileName, feAuto, Encoding);
  {$IFDEF FULL_PREPROCESSOR}
    // Convert all relative paths to absolute
    Result := '';
    Paths := RegExCreate('(?P<PREFIX><[^>]*?\s(?:href|src)=(?P<QUOTE>["'']?))(?P<SCHEMA>[a-z]+:)?(?P<PATH>(?!#)[^>]*?)(?P<SUFFIX>(?P=QUOTE))');
    try
      if Paths.Match(Content) then
        repeat
          // Only convert paths with no schema (protocol)
          if (Paths.SubexpressionByName('SCHEMA') = '') then
            begin
            Path := Paths.SubexpressionByName('PATH');
            // If there's a path, convert it
            if Path <> '' then
              begin
              Paths.Replacement := ''
                + '${PREFIX}' +
                {$IFDEF UNICODE} PCREString {$ENDIF} (
                  AbsoluteLocalUrlPrefix + StringReplace(StringReplace(RelativePathToFullPath(FileName, Paths.SubExpressionByName('PATH')), '\', '/', [rfReplaceAll]), '$', '\$', [rfReplaceAll])
                ) + '${SUFFIX}';
              Paths.Replace;
              end;
            end;
        until not Paths.MatchAgain;
      Result := {$IFDEF UNICODE} string {$ENDIF} (Paths.Subject);
    finally
      Paths.Free;
      end;
    Content := Result;
  {$ENDIF}
  // Process all include tags
  Result := '';
  Includes := RegExCreate('<meta\s+name=(?P<QUOTE1>["'']?)(?P<NAME>[^>]*?)(?P=QUOTE1)\s+content=(?P<QUOTE2>["'']?)(?P<VALUE>[^>]*?)(?P=QUOTE2).*?>');
  try
    if Includes.Match(Content) then
      repeat
        Name := DecodeHtmlString(Includes.SubexpressionByName('NAME'));
        Value := DecodeHtmlString(Includes.SubexpressionByName('VALUE'));
        ix := Pos(UpperCase(META_INCLUDE), UpperCase(Name));
        if (ix = 1) or ((ix > 1) and (Name[ix-1] = ':')) then
          begin
          if ValidMetaInclude(Name) then
            Includes.Replacement :=
              {$IFDEF UNICODE} PCREString {$ENDIF} (
                StringReplace(
                  StringReplace(
                    ReadFileWithAllIncludes(RelativePathToFullPath(FileName, Value), DummyEncoding),
                    '\', '\\', [rfReplaceAll]
                  ), '$', '\$', [rfReplaceAll]
                )
              )
          else
            Includes.Replacement := '';
          Includes.Replace;
          end;
      until not Includes.MatchAgain;
    Result := {$IFDEF UNICODE} string {$ENDIF} (Includes.Subject);
  finally
    Includes.Free;
    end;
end;

{$IFDEF FULL_PREPROCESSOR}
function StoreContentToTempFile(const FileName, Content: string; Encoding: TFileEncoding): string;
  // Create a new temp file and store the content into it
var
  Dir: string;
  {$IFNDEF UNICODE}
  ContentW: WideString;
  {$ENDIF}
begin
  Dir := SystemTempFile(SystemTempDir, 'h2lrf');
  if FileExists(Dir) then
    SysUtils.DeleteFile(Dir);
  ForceDirectories(Dir);
  {$IFDEF CLEANUP}
    if CleanUp <> nil then
      CleanUp.Add(Dir);
  {$ENDIF}
  Result := Dir + '\' + ExtractFileName(FileName);
  {$IFDEF UNICODE}
    SaveMemoryToFileW(Result, @Content[1], Length(Content)*Sizeof(Char), Encoding);
  {$ELSE}
    ContentW := Content;
    SaveMemoryToFileW(Result, @ContentW[1], Length(ContentW)*Sizeof(WideChar), Encoding);
  {$ENDIF}
  {$IFDEF LOG_OUTPUT_FILE}
    {$IFDEF UNICODE}
      SaveMemoryToFileW(ExtractFileName(Result), @Content[1], Length(Content)*Sizeof(Char), Encoding);
    {$ELSE}
      SaveMemoryToFileW(ExtractFileName(Result), @ContentW[1], Length(ContentW)*Sizeof(WideChar), Encoding);
    {$ENDIF}
  {$ENDIF}
end;
{$ENDIF}

function CreateCommandLine(const RootDir, FileName, Extension, OutputDir, CommandLinePattern: string; RecursiveOutput: boolean;  out CommandLine: string): boolean;
  // Create command line for converter
var Params, Values: TStringList;
    CLInput, CLOutput, CLParams: string;
    s, Pattern: string;
    FN, Obsah, Name, Value: string;
    i: integer;
    Encoding: TFileEncoding;
begin
  Result := False;
  FN := RootDir + FileName;
  // Nactu obsah souboru a vsech jeho includovanych souboru
  Obsah := ReadFileWithAllIncludes(FN, Encoding);
  // Nactu si z nej parametry
  Params := TStringList.Create;
  try
    Values := TStringList.Create;
    try
      if ReadParametersFromFile(Obsah, FN, Params, Values) then
        begin
        // Tim mam nactene parametry, muzu se pustit do vytvoreni prikazove radky
        SetLength(CommandLine, 2048);
        CommandLine := '';
        // Jmeno vstupniho souboru
        CLInput := {$IFDEF FULL_PREPROCESSOR} StoreContentToTempFile(FileName, Obsah, Encoding) {$ELSE} RootDir + FileName {$ENDIF};
        // Jmeno vystupniho souboru
        if RecursiveOutput then
          CLOutput := IncludeTrailingBackslash(OutputDir) + ChangeFileExt(FileName, Extension)
        else
          CLOutput := IncludeTrailingBackslash(OutputDir) + ChangeFileExt(ExtractFileName(FileName), Extension);
        ForceDirectories(ExtractFileDir(CLOutput));
        // Parametry
        CLParams := '';
        for i := Pred(Params.Count) downto 0 do
          begin
          Name := Params[i];
          Value := Values[i];
          if Name <> '' then
            begin
            if Value <> '' then
              if Pos(' ', Value) > 0 then
                CLParams := Name + '=' + EscapeQuotes(Value) + ' ' + CLParams
              else
                CLParams := Name + '=' + Value + ' ' + CLParams
            else
              CLParams := Name + ' ' + CLParams;
            end;
          end;
        // Ted to cele poskladam dohromady
        Pattern := CommandLinePattern;
        repeat
          i := Pos('%', Pattern);
          if i > 0 then
            begin
            CommandLine := CommandLine + Copy(Pattern, 1, Pred(i));
            System.Delete(Pattern, 1, i);
            i := Pos('%', Pattern);
            if i <= 0 then
              CommandLine := CommandLine + '%'
            else if i = 1 then
              begin
              CommandLine := CommandLine + '%';
              System.Delete(Pattern, 1, 1);
              end
            else
              begin
              s := UpperCase(Copy(Pattern, 1, Pred(i)));
              System.Delete(Pattern, 1, i);
              if s = 'PARAMS' then
                CommandLine := CommandLine + CLParams
              else if s = 'INPUT' then
                CommandLine := CommandLine + CLInput
              else if s = 'OUTPUT' then
                CommandLine := CommandLine + CLOutput
              else
                begin
                CommandLine := CommandLine + s;
                Pattern := '%' + Pattern;
                end;
              end;
            end
          else
            Break;
        until False;
        CommandLine := CommandLine + Pattern;
        Result := True;
        end;
    finally
      Values.Free;
      end;
  finally
    Params.Free;
    end;
end;

function Convert(const Options: TConvertOptions): boolean;
  // Convert file(s) to the new format
var SR: TSearchRec;
    OK: integer;
    Dir, Mask: string;
    CommandLine: string;
    ResultCode: DWORD;
    Opt: TConvertOptions;
begin
  MetaPrefix := Options.InputMetaPrefix;
  MetaPrefixLength := Length(MetaPrefix);
  AbsoluteLocalUrlPrefix := Options.AbsoluteLocalUrlPrefix;
  OK := SysUtils.FindFirst(Options.InputDir + Options.InputWildCard, faAnyFile, SR);
  if OK = 0 then
    try
      Result := True;
      Dir := ExtractFilePath(Options.InputWildCard);
      Mask := ExtractFileName(Options.InputWildCard);
      while OK = 0 do
        begin
        if Longbool(SR.Attr and faDirectory) then
          begin
          // Nothing to do
          end
        else
          try
            if CreateCommandLine(Options.InputDir, Dir + SR.Name, Options.OutputExtension, Options.OutputDir, Options.ConverterCommandLine, Options.OutputRecursive, CommandLine) then
              begin
              if Execute(Options.ConverterPath, CommandLine, ResultCode) then
                begin
                if ResultCode <> 0 then
                  begin
                  Result := False;
                  Raise Eh2lrfError.CreateFmt('%s returned errorlevel %d.%sCommand Line: %s %s', [ExtractFileName(Options.ConverterPath), ResultCode, EOLN, Options.ConverterPath, CommandLine]);
                  end;
                end
              else
                begin
                Result := False;
                ResultCode := {$IFDEF FPC} DosError {$ELSE} GetLastError {$ENDIF};
                Raise Eh2lrfError.CreateFmt('Could not create process: (%x) %s.%sCommand Line: %s %s', [ResultCode, GetLastErrorStr(ResultCode), EOLN, Options.ConverterPath, CommandLine]);
                end;
              end
            else
              Result := False;
            {$IFDEF CLEANUP}
              DoCleanUp;
            {$ENDIF}
          except
            on E: Exception do
              begin
              ReportError(E, Options.ErrorLogFileName, Options.InputDir + Dir + SR.Name);
              if Options.ErrorLogFileName = '' then
                Raise;
              end;
          end;
        OK := SysUtils.FindNext(SR);
        end;
    finally
      SysUtils.FindClose(SR);
      end
  else
    Result := False;
  if Options.InputRecursive then
    begin
    Dir := ExtractFilePath(Options.InputWildCard);
    Mask := ExtractFileName(Options.InputWildCard);
    Opt := Options;
    OK := SysUtils.FindFirst(Options.InputDir + Dir + '*.*', faDirectory, SR);
    if OK = 0 then
      try
        while OK = 0 do
          begin
          if Longbool(SR.Attr and faDirectory) and (SR.Name <> '.') and (SR.Name <> '..') then
            begin
            Opt.InputWildCard := Dir + SR.Name + '\' + Mask;
            if not Convert(Opt) then
              Result := False;
            end;
          OK := SysUtils.FindNext(SR);
          end;
      finally
        SysUtils.FindClose(SR);
        end;
    end;
end;

procedure Initialize;
begin
  {$IFDEF CLEANUP}
    CleanUp := TStringList.Create;
  {$ENDIF}
  MetaPrefix := DEFAULT_TARGET + ':';
  MetaPrefixLength := Length(MetaPrefix);
end;

procedure Finalize;
begin
  {$IFDEF CLEANUP}
    DoCleanUp;
    CleanUp.Free;
    CleanUp := nil;
  {$ENDIF}
end;

initialization
  Initialize;

finalization
  Finalize;

end.
